import { mergeBytes, isSome, none, some, utf8, fixSerializer, publicKey } from '@metaplex-foundation/umi';
import { DataViewSerializerError, DeserializingEmptyBufferError, NotEnoughBytesError } from './errors.mjs';
import { u8, u16, u32, u64, u128, i8, i16, i32, i64, i128, f32, f64 } from './numbers.mjs';

function createDataViewSerializer(options = {}) {
  const tuple = (items, options = {}) => {
    const itemDescriptions = items.map(item => item.description).join(', ');
    return {
      description: options.description ?? `tuple(${itemDescriptions})`,
      fixedSize: sumSerializerSizes(items.map(item => item.fixedSize)),
      maxSize: sumSerializerSizes(items.map(item => item.maxSize)),
      serialize: value => {
        if (value.length !== items.length) {
          throw new DataViewSerializerError(`Expected tuple to have ${items.length} items but got ${value.length}.`);
        }
        return mergeBytes(items.map((item, index) => item.serialize(value[index])));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        items.forEach(serializer => {
          const [newValue, newOffset] = serializer.deserialize(bytes, offset);
          values.push(newValue);
          offset = newOffset;
        });
        return [values, offset];
      }
    };
  };
  const array = (item, options = {}) => {
    const size = options.size ?? u32();
    if (size === 'remainder' && item.fixedSize === null) {
      throw new DataViewSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    return {
      description: options.description ?? `array(${item.description}; ${getSizeDescription(size)})`,
      fixedSize: getSizeFromChildren(size, [item.fixedSize]),
      maxSize: getSizeFromChildren(size, [item.maxSize]),
      serialize: value => {
        if (typeof size === 'number' && value.length !== size) {
          throw new DataViewSerializerError(`Expected array to have ${size} items but got ${value.length}.`);
        }
        return mergeBytes([getSizePrefix(size, value.length), ...value.map(v => item.serialize(v))]);
      },
      deserialize: (bytes, offset = 0) => {
        if (typeof size === 'object' && bytes.slice(offset).length === 0) {
          return handleEmptyBuffer('array', [], offset);
        }
        const [resolvedSize, newOffset] = getResolvedSize(size, [item.fixedSize], bytes, offset);
        offset = newOffset;
        const values = [];
        for (let i = 0; i < resolvedSize; i += 1) {
          const [value, newOffset] = item.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  };
  const map = (key, value, options = {}) => {
    const size = options.size ?? u32();
    if (size === 'remainder' && (key.fixedSize === null || value.fixedSize === null)) {
      throw new DataViewSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    return {
      description: options.description ?? `map(${key.description}, ${value.description}; ${getSizeDescription(size)})`,
      fixedSize: getSizeFromChildren(size, [key.fixedSize, value.fixedSize]),
      maxSize: getSizeFromChildren(size, [key.maxSize, value.maxSize]),
      serialize: map => {
        if (typeof size === 'number' && map.size !== size) {
          throw new DataViewSerializerError(`Expected map to have ${size} items but got ${map.size}.`);
        }
        const itemBytes = Array.from(map, ([k, v]) => mergeBytes([key.serialize(k), value.serialize(v)]));
        return mergeBytes([getSizePrefix(size, map.size), ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const map = new Map();
        if (typeof size === 'object' && bytes.slice(offset).length === 0) {
          return handleEmptyBuffer('map', map, offset);
        }
        const [resolvedSize, newOffset] = getResolvedSize(size, [key.fixedSize, value.fixedSize], bytes, offset);
        offset = newOffset;
        for (let i = 0; i < resolvedSize; i += 1) {
          const [deserializedKey, kOffset] = key.deserialize(bytes, offset);
          offset = kOffset;
          const [deserializedValue, vOffset] = value.deserialize(bytes, offset);
          offset = vOffset;
          map.set(deserializedKey, deserializedValue);
        }
        return [map, offset];
      }
    };
  };
  const set = (item, options = {}) => {
    const size = options.size ?? u32();
    if (size === 'remainder' && item.fixedSize === null) {
      throw new DataViewSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    return {
      description: options.description ?? `set(${item.description}; ${getSizeDescription(size)})`,
      fixedSize: getSizeFromChildren(size, [item.fixedSize]),
      maxSize: getSizeFromChildren(size, [item.maxSize]),
      serialize: set => {
        if (typeof size === 'number' && set.size !== size) {
          throw new DataViewSerializerError(`Expected set to have ${size} items but got ${set.size}.`);
        }
        const itemBytes = Array.from(set, value => item.serialize(value));
        return mergeBytes([getSizePrefix(size, set.size), ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const set = new Set();
        if (typeof size === 'object' && bytes.slice(offset).length === 0) {
          return handleEmptyBuffer('set', set, offset);
        }
        const [resolvedSize, newOffset] = getResolvedSize(size, [item.fixedSize], bytes, offset);
        offset = newOffset;
        for (let i = 0; i < resolvedSize; i += 1) {
          const [value, newOffset] = item.deserialize(bytes, offset);
          offset = newOffset;
          set.add(value);
        }
        return [set, offset];
      }
    };
  };
  const option = (item, options = {}) => {
    const prefix = options.prefix ?? u8();
    const fixed = options.fixed ?? false;
    let descriptionSuffix = `; ${getSizeDescription(prefix)}`;
    let fixedSize = item.fixedSize === 0 ? prefix.fixedSize : null;
    if (fixed) {
      if (item.fixedSize === null || prefix.fixedSize === null) {
        throw new DataViewSerializerError('Fixed options can only be used with fixed-size serializers');
      }
      descriptionSuffix += '; fixed';
      fixedSize = prefix.fixedSize + item.fixedSize;
    }
    return {
      description: options.description ?? `option(${item.description + descriptionSuffix})`,
      fixedSize,
      maxSize: sumSerializerSizes([prefix.maxSize, item.maxSize]),
      serialize: option => {
        const prefixByte = prefix.serialize(Number(isSome(option)));
        if (fixed) {
          const itemFixedSize = item.fixedSize;
          const itemBytes = isSome(option) ? item.serialize(option.value).slice(0, itemFixedSize) : new Uint8Array(itemFixedSize).fill(0);
          return mergeBytes([prefixByte, itemBytes]);
        }
        const itemBytes = isSome(option) ? item.serialize(option.value) : new Uint8Array();
        return mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          return handleEmptyBuffer('option', none(), offset);
        }
        const fixedOffset = offset + (prefix.fixedSize ?? 0) + (item.fixedSize ?? 0);
        const [isSome, prefixOffset] = prefix.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [none(), fixed ? fixedOffset : offset];
        }
        const [value, newOffset] = item.deserialize(bytes, offset);
        offset = newOffset;
        return [some(value), fixed ? fixedOffset : offset];
      }
    };
  };
  const nullable = (item, options = {}) => {
    const prefix = options.prefix ?? u8();
    const fixed = options.fixed ?? false;
    let descriptionSuffix = `; ${getSizeDescription(prefix)}`;
    let fixedSize = item.fixedSize === 0 ? prefix.fixedSize : null;
    if (fixed) {
      if (item.fixedSize === null || prefix.fixedSize === null) {
        throw new DataViewSerializerError('Fixed nullables can only be used with fixed-size serializers');
      }
      descriptionSuffix += '; fixed';
      fixedSize = prefix.fixedSize + item.fixedSize;
    }
    return {
      description: options.description ?? `nullable(${item.description + descriptionSuffix})`,
      fixedSize,
      maxSize: sumSerializerSizes([prefix.maxSize, item.maxSize]),
      serialize: option => {
        const prefixByte = prefix.serialize(Number(option !== null));
        if (fixed) {
          const itemFixedSize = item.fixedSize;
          const itemBytes = option !== null ? item.serialize(option).slice(0, itemFixedSize) : new Uint8Array(itemFixedSize).fill(0);
          return mergeBytes([prefixByte, itemBytes]);
        }
        const itemBytes = option !== null ? item.serialize(option) : new Uint8Array();
        return mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          return handleEmptyBuffer('nullable', null, offset);
        }
        const fixedOffset = offset + (prefix.fixedSize ?? 0) + (item.fixedSize ?? 0);
        const [isSome, prefixOffset] = prefix.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [null, fixed ? fixedOffset : offset];
        }
        const [value, newOffset] = item.deserialize(bytes, offset);
        offset = newOffset;
        return [value, fixed ? fixedOffset : offset];
      }
    };
  };
  const struct = (fields, options = {}) => {
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}: ${serializer.description}`).join(', ');
    return {
      description: options.description ?? `struct(${fieldDescriptions})`,
      fixedSize: sumSerializerSizes(fields.map(([, field]) => field.fixedSize)),
      maxSize: sumSerializerSizes(fields.map(([, field]) => field.maxSize)),
      serialize: struct => {
        const fieldBytes = fields.map(([key, serializer]) => serializer.serialize(struct[key]));
        return mergeBytes(fieldBytes);
      },
      deserialize: (bytes, offset = 0) => {
        const struct = {};
        fields.forEach(([key, serializer]) => {
          const [value, newOffset] = serializer.deserialize(bytes, offset);
          offset = newOffset;
          struct[key] = value;
        });
        return [struct, offset];
      }
    };
  };
  const enumFn = (constructor, options = {}) => {
    const prefix = options.size ?? u8();
    const enumKeys = Object.keys(constructor);
    const enumValues = Object.values(constructor);
    const isNumericEnum = enumValues.some(v => typeof v === 'number');
    const valueDescriptions = enumValues.filter(v => typeof v === 'string').join(', ');
    const minRange = 0;
    const maxRange = isNumericEnum ? enumValues.length / 2 - 1 : enumValues.length - 1;
    const stringValues = isNumericEnum ? [...enumKeys] : [...new Set([...enumKeys, ...enumValues])];
    function assertValidVariant(variant) {
      const isInvalidNumber = typeof variant === 'number' && (variant < minRange || variant > maxRange);
      const isInvalidString = typeof variant === 'string' && !stringValues.includes(variant);
      if (isInvalidNumber || isInvalidString) {
        throw new DataViewSerializerError(`Invalid enum variant. Got "${variant}", ` + `expected one of [${stringValues.join(', ')}] ` + `or a number between ${minRange} and ${maxRange}`);
      }
    }
    return {
      description: options.description ?? `enum(${valueDescriptions}; ${prefix.description})`,
      fixedSize: prefix.fixedSize,
      maxSize: prefix.maxSize,
      serialize: value => {
        assertValidVariant(value);
        if (typeof value === 'number') return prefix.serialize(value);
        const valueIndex = enumValues.indexOf(value);
        if (valueIndex >= 0) return prefix.serialize(valueIndex);
        return prefix.serialize(enumKeys.indexOf(value));
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new DeserializingEmptyBufferError('enum');
        }
        const [value, newOffset] = prefix.deserialize(bytes, offset);
        const valueAsNumber = Number(value);
        offset = newOffset;
        assertValidVariant(valueAsNumber);
        return [isNumericEnum ? valueAsNumber : enumValues[valueAsNumber], offset];
      }
    };
  };
  const dataEnum = (variants, options = {}) => {
    const prefix = options.size ?? u8();
    const fieldDescriptions = variants.map(([name, serializer]) => `${String(name)}${serializer ? `: ${serializer.description}` : ''}`).join(', ');
    const allVariantHaveTheSameFixedSize = variants.every((one, i, all) => one[1].fixedSize === all[0][1].fixedSize);
    const fixedVariantSize = allVariantHaveTheSameFixedSize ? variants[0][1].fixedSize : null;
    const maxVariantSize = maxSerializerSizes(variants.map(([, field]) => field.maxSize));
    return {
      description: options.description ?? `dataEnum(${fieldDescriptions}; ${prefix.description})`,
      fixedSize: variants.length === 0 ? prefix.fixedSize : sumSerializerSizes([prefix.fixedSize, fixedVariantSize]),
      maxSize: variants.length === 0 ? prefix.maxSize : sumSerializerSizes([prefix.maxSize, maxVariantSize]),
      serialize: variant => {
        const discriminator = variants.findIndex(([key]) => variant.__kind === key);
        if (discriminator < 0) {
          throw new DataViewSerializerError(`Invalid data enum variant. Got "${variant.__kind}", expected one of ` + `[${variants.map(([key]) => key).join(', ')}]`);
        }
        const variantPrefix = prefix.serialize(discriminator);
        const variantSerializer = variants[discriminator][1];
        const variantBytes = variantSerializer.serialize(variant);
        return mergeBytes([variantPrefix, variantBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new DeserializingEmptyBufferError('dataEnum');
        }
        const [discriminator, dOffset] = prefix.deserialize(bytes, offset);
        offset = dOffset;
        const variantField = variants[Number(discriminator)] ?? null;
        if (!variantField) {
          throw new DataViewSerializerError(`Data enum index "${discriminator}" is out of range. ` + `Index should be between 0 and ${variants.length - 1}.`);
        }
        const [variant, vOffset] = variantField[1].deserialize(bytes, offset);
        offset = vOffset;
        return [{
          __kind: variantField[0],
          ...(variant ?? {})
        }, offset];
      }
    };
  };
  const string = (options = {}) => {
    const size = options.size ?? u32();
    const encoding = options.encoding ?? utf8;
    const description = options.description ?? `string(${encoding.description}; ${getSizeDescription(size)})`;
    if (size === 'variable') {
      return {
        ...encoding,
        description
      };
    }
    if (typeof size === 'number') {
      return fixSerializer(encoding, size, description);
    }
    return {
      description,
      fixedSize: null,
      maxSize: null,
      serialize: value => {
        const contentBytes = encoding.serialize(value);
        const lengthBytes = size.serialize(contentBytes.length);
        return mergeBytes([lengthBytes, contentBytes]);
      },
      deserialize: (buffer, offset = 0) => {
        if (buffer.slice(offset).length === 0) {
          throw new DeserializingEmptyBufferError('string');
        }
        const [lengthBigInt, lengthOffset] = size.deserialize(buffer, offset);
        const length = Number(lengthBigInt);
        offset = lengthOffset;
        const contentBuffer = buffer.slice(offset, offset + length);
        if (contentBuffer.length < length) {
          throw new NotEnoughBytesError('string', length, contentBuffer.length);
        }
        const [value, contentOffset] = encoding.deserialize(contentBuffer);
        offset += contentOffset;
        return [value, offset];
      }
    };
  };
  const bool = (options = {}) => {
    const size = options.size ?? u8();
    if (size.fixedSize === null) {
      throw new DataViewSerializerError('Serializer [bool] requires a fixed size.');
    }
    return {
      description: options.description ?? `bool(${size.description})`,
      fixedSize: size.fixedSize,
      maxSize: size.fixedSize,
      serialize: value => size.serialize(value ? 1 : 0),
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new DeserializingEmptyBufferError('bool');
        }
        const [value, vOffset] = size.deserialize(bytes, offset);
        return [value === 1, vOffset];
      }
    };
  };
  const unit = (options = {}) => ({
    description: options.description ?? 'unit',
    fixedSize: 0,
    maxSize: 0,
    serialize: () => new Uint8Array(),
    deserialize: (_bytes, offset = 0) => [undefined, offset]
  });
  const bytes = (options = {}) => {
    const size = options.size ?? 'variable';
    const description = options.description ?? `bytes(${getSizeDescription(size)})`;
    const byteSerializer = {
      description,
      fixedSize: null,
      maxSize: null,
      serialize: value => new Uint8Array(value),
      deserialize: (bytes, offset = 0) => {
        const slice = new Uint8Array(bytes.slice(offset));
        return [slice, offset + slice.length];
      }
    };
    if (size === 'variable') {
      return byteSerializer;
    }
    if (typeof size === 'number') {
      return fixSerializer(byteSerializer, size, description);
    }
    return {
      description,
      fixedSize: null,
      maxSize: null,
      serialize: value => {
        const contentBytes = byteSerializer.serialize(value);
        const lengthBytes = size.serialize(contentBytes.length);
        return mergeBytes([lengthBytes, contentBytes]);
      },
      deserialize: (buffer, offset = 0) => {
        if (buffer.slice(offset).length === 0) {
          throw new DeserializingEmptyBufferError('bytes');
        }
        const [lengthBigInt, lengthOffset] = size.deserialize(buffer, offset);
        const length = Number(lengthBigInt);
        offset = lengthOffset;
        const contentBuffer = buffer.slice(offset, offset + length);
        if (contentBuffer.length < length) {
          throw new NotEnoughBytesError('bytes', length, contentBuffer.length);
        }
        const [value, contentOffset] = byteSerializer.deserialize(contentBuffer);
        offset += contentOffset;
        return [value, offset];
      }
    };
  };
  const publicKeyFn = (options = {}) => ({
    description: options.description ?? 'publicKey',
    fixedSize: 32,
    maxSize: 32,
    serialize: value => publicKey(value).bytes,
    deserialize: (bytes, offset = 0) => {
      if (bytes.slice(offset).length === 0) {
        throw new DeserializingEmptyBufferError('publicKey');
      }
      const pubkeyBytes = bytes.slice(offset, offset + 32);
      if (pubkeyBytes.length < 32) {
        throw new DataViewSerializerError(`Serializer [publicKey] expected 32 bytes, got ${pubkeyBytes.length}.`);
      }
      return [publicKey(pubkeyBytes), offset + 32];
    }
  });
  const handleEmptyBuffer = (serializer, defaultValue, offset) => {
    if (!(options.tolerateEmptyBuffers ?? true)) {
      throw new DeserializingEmptyBufferError(serializer);
    }
    return [defaultValue, offset];
  };
  return {
    tuple,
    array,
    map,
    set,
    option,
    nullable,
    struct,
    enum: enumFn,
    dataEnum,
    string,
    bool,
    unit,
    u8,
    u16,
    u32,
    u64,
    u128,
    i8,
    i16,
    i32,
    i64,
    i128,
    f32,
    f64,
    bytes,
    publicKey: publicKeyFn
  };
}
function sumSerializerSizes(sizes) {
  return sizes.reduce((all, size) => all === null || size === null ? null : all + size, 0);
}
function maxSerializerSizes(sizes) {
  return sizes.reduce((all, size) => all === null || size === null ? null : Math.max(all, size), 0);
}
function getSizeDescription(size) {
  return typeof size === 'object' ? size.description : `${size}`;
}
function getSizeFromChildren(size, childrenSizes) {
  if (typeof size !== 'number') return null;
  if (size === 0) return 0;
  const childrenSize = sumSerializerSizes(childrenSizes);
  return childrenSize === null ? null : childrenSize * size;
}
function getSizePrefix(size, realSize) {
  return typeof size === 'object' ? size.serialize(realSize) : new Uint8Array();
}
function getResolvedSize(size, childrenSizes, bytes, offset) {
  if (typeof size === 'number') {
    return [size, offset];
  }
  if (typeof size === 'object') {
    return size.deserialize(bytes, offset);
  }
  if (size === 'remainder') {
    const childrenSize = sumSerializerSizes(childrenSizes);
    if (childrenSize === null) {
      throw new DataViewSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    const remainder = bytes.slice(offset).length;
    if (remainder % childrenSize !== 0) {
      throw new DataViewSerializerError(`Serializers of "remainder" size must have a remainder that is a multiple of its item size. ` + `Got ${remainder} bytes remaining and ${childrenSize} bytes per item.`);
    }
    return [remainder / childrenSize, offset];
  }
  throw new DataViewSerializerError(`Unknown size type: ${JSON.stringify(size)}.`);
}

export { createDataViewSerializer };
//# sourceMappingURL=createDataViewSerializer.mjs.map
